@@ -9,6 +9,7 @@ from six import text_type, binary_type, integer_types
 import string
 import random as r
 import hypothesis.descriptors as descriptors
+from copy import deepcopy
 
 
 def mix_generators(*generators):
@@ -41,6 +42,8 @@ def nice_string(xs):
                 map(nice_string, xs)
             )
         )
+    if isinstance(xs, frozenset):
+        return "frozenset(%s)" % (nice_string(set(xs)),)
     try:
         return xs.__name__
     except AttributeError:
@@ -60,6 +63,7 @@ def nice_string(xs):
 
 
 class SearchStrategy(object):
+    has_immutable_data = True
 
     def __repr__(self):
         return '%s(%s)' % (
@@ -74,10 +78,17 @@ class SearchStrategy(object):
     def produce(self, random, parameter_value):
         pass  # pragma: no cover
 
+    def copy(self, value):
+        if self.has_immutable_data:
+            return value
+        else:
+            return deepcopy(value)
+
     def simplify(self, value):
         return iter(())
 
     def simplify_such_that(self, t, f):
+        assert self.could_have_produced(t)
         if not f(t):
             raise ValueError(
                 "%r does not satisfy predicate %s" % (t, f))
@@ -86,6 +97,7 @@ class SearchStrategy(object):
 
         while True:
             for s in self.simplify(t):
+                assert self.could_have_produced(s)
                 if tracker.track(s) > 1:
                     continue
                 if f(s):
@@ -285,6 +297,7 @@ class TupleStrategy(SearchStrategy):
         self.parameter = params.CompositeParameter(
             x.parameter for x in self.element_strategies
         )
+        self.has_immutable_data = all(s.has_immutable_data for s in strategies)
 
     def could_have_produced(self, xs):
         if xs.__class__ != self.tuple_type:
@@ -351,6 +364,7 @@ def _unique(xs):
 
 
 class ListStrategy(SearchStrategy):
+    has_immutable_data = False
 
     def __init__(self,
                  strategies, average_length=100.0):
@@ -378,7 +392,7 @@ class ListStrategy(SearchStrategy):
 
         generators.append(iter(([],)))
 
-        indices = xrange(0, len(x))
+        indices = xrange(len(x) - 1, -1, -1)
 
         generators.append(
             [x[i]] for i in indices
@@ -472,6 +486,7 @@ class ComplexStrategy(SearchStrategy):
 
 
 class SetStrategy(MappedSearchStrategy):
+    has_immutable_data = False
 
     def __init__(self, list_strategy):
         super(SetStrategy, self).__init__(
@@ -486,6 +501,23 @@ class SetStrategy(MappedSearchStrategy):
         return list(x)
 
 
+class FrozenSetStrategy(MappedSearchStrategy):
+    def __init__(self, list_strategy):
+        super(FrozenSetStrategy, self).__init__(
+            strategy=list_strategy,
+            descriptor=frozenset(list_strategy.descriptor)
+        )
+        self.has_immutable_data = (
+            list_strategy.element_strategy.has_immutable_data
+        )
+
+    def pack(self, x):
+        return frozenset(x)
+
+    def unpack(self, x):
+        return list(x)
+
+
 class OneCharStringStrategy(SearchStrategy):
     descriptor = text_type
 
@@ -530,6 +562,7 @@ class BinaryStringStrategy(MappedSearchStrategy):
 
 
 class FixedKeysDictStrategy(SearchStrategy):
+    has_immutable_data = False
 
     def __init__(self, strategy_dict):
         SearchStrategy.__init__(self)
@@ -582,6 +615,8 @@ class OneOfStrategy(SearchStrategy):
                 e.parameter for e in self.element_strategies
             )
         )
+        self.has_immutable_data = all(
+            s.has_immutable_data for s in self.element_strategies)
 
     def could_have_produced(self, x):
         return any((s.could_have_produced(x) for s in self.element_strategies))
@@ -605,6 +640,11 @@ class OneOfStrategy(SearchStrategy):
 
 
 class JustStrategy(SearchStrategy):
+    # We could do better here but it's probably not worth it
+    # deepcopy has optimisations that will probably work just as well as
+    # our check
+    has_immutable_data = False
+
     def __init__(self, value):
         self.descriptor = descriptors.Just(value)
 
@@ -612,3 +652,42 @@ class JustStrategy(SearchStrategy):
 
     def produce(self, random, pv):
         return self.descriptor.value
+
+    def could_have_produced(self, value):
+        return self.descriptor.value == value
+
+
+class RandomWithSeed(r.Random):
+    def __init__(self, seed):
+        super(RandomWithSeed, self).__init__(seed)
+        self.seed = seed
+
+    def __repr__(self):
+        return "Random(%s)" % (self.seed,)
+
+    def __copy__(self):
+        r = RandomWithSeed(self.seed)
+        r.setstate(self.getstate())
+        return r
+
+    def __deepcopy__(self, d):
+        return self.__copy__()
+
+    def __eq__(self, other):
+        return self is other or (
+            isinstance(other, RandomWithSeed) and
+            self.seed == other.seed and
+            self.getstate() == other.getstate()
+        )
+
+
+class RandomStrategy(SearchStrategy):
+    descriptor = r.Random
+    parameter = params.CompositeParameter()
+    has_immutable_data = False
+
+    def produce(self, random, pv):
+        return RandomWithSeed(random.getrandbits(128))
+
+    def could_have_produced(self, value):
+        return isinstance(value, RandomWithSeed)
